//----------------------------------------------------------------------
//- Resplendent Filefind Component v 2.00 - Nov 1998
//- (c) Copyright 1998 Resplendence Sp
//----------------------------------------------------------------------

unit RFilefind;

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, FileCtrl;

const
   bufsize=32767;
   versionStr='2.00';

type
  TChangeDirEvent=procedure(Sender:TObject; directory: String) of object;
  TMatchEvent=procedure(Sender:TObject; directory: String; fileinfo: TSearchRec) of object;
  TTextMatchEvent=procedure(Sender:TObject; directory: String; fileinfo: TSearchRec; offset: Integer; var processNextMatch: Boolean) of object;
  TScanEvent=procedure(Sender:TObject; directory: String; fileinfo: TSearchRec; var AllowScanning: Boolean) of Object;

  TSearchState=(ssInitializing, ssRunning, ssPaused, ssStopped, ssCompleted);

  TFindThread = class(TThread)
  private
    ContinueProcessing,AllowProcessing: Boolean;
    TextLength: Integer;
    foundrec: TSearchRec;
    currentOffset: Integer;
    procedure RecursiveSearchDir(dir: String);
    procedure TestIfFileInDirectory (dir: String);
    procedure TestIfFileContainsText;
    { Private declarations }
    procedure MatchFound;
    procedure TextFound;
    procedure WhenCompleted;
    procedure WhenDirChange;
    procedure WhenDenied;
    procedure WhenScanning;
  protected
     procedure Execute; override;
  public
    CaseSensitive: Boolean;
    FoundText: TTextMatchEvent;
    Found,Denied: TMatchEvent;
    Scanning: TScanEvent;
    DirChange: TChangeDirEvent;
    Completed: TNotifyEvent;
    Text,filemask,directory: String;
    Files: TStrings;
    Paths: TStrings;
    procedure Stop;
    constructor Create;
    destructor Destroy; override;
  end;

  TRFilefind = class(TComponent)
  private
    _Directories,_Files: TStrings;
    _Threadpriority: TThreadPriority;
    _caseSensitive: Boolean;
    _onAccessDenied: TMatchEvent;
    _onChangeDir: TChangeDirEvent;
    _onCompleted: TNotifyEvent;
    _onFileFound: TMatchEvent;
    _onTextFound: TTextMatchEvent;
    _onStartScanning: TScanEvent;
    _text,_filemask: String;
    _state: TSearchState;
    _version: String;

    Thread: TFindThread;
    procedure Set_Directories  (list: TStrings);
    procedure Set_Files        (list: TStrings);
    procedure Set_Text         (str: String);
    procedure Set_Version      (str: String);

  protected
    procedure Completed (Sender: TObject);
    //
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function    GetCDRomDrives:    String;
    function    GetFixedDrives:     String;
    function    GetRamDrives:     String;
    function    GetRemoteDrives:    String;
    function    GetRemovableDrives: String;
    procedure   Continue;
    procedure   Pause;
    procedure   Start;
    procedure   Stop;


    //
  published
    property OnAccessDenied: TMatchEvent read _onAccessDenied write _onAccessDenied;
    property OnChangeDirectory: TChangeDirEvent read _onChangeDir write _onChangeDir;
    property OnCompleted: TNotifyEvent read _onCompleted write _onCompleted;
    property OnFileFound: TMatchEvent  read _onFileFound write _onFileFound;
    property OnTextFound: TTextMatchEvent  read _onTextFound write _onTextFound;
    property OnStartScanning: TScanEvent read _onStartScanning write _onStartScanning;

    property CaseSensitive:  Boolean         read _caseSensitive   write _caseSensitive;
    property ContainingText: String          read _text            write Set_Text;
    property Directories:    TStrings        read _Directories     write Set_Directories;
    property FileMask:       String          read _filemask        write _FileMask;
    property Files:          TStrings        read _files           write Set_Files;
    property ThreadPriority: TThreadPriority read _ThreadPriority  write _ThreadPriority;
    property State:          TSearchState    read _state;
    property Version:        String          read _version         write Set_Version;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Resplendence', [TRFilefind]);
end;

constructor TRFilefind.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);

    _version:=versionStr;
    _ThreadPriority:=tpNormal;

    _filemask:='*.exe';

    _Directories:=TStringList.Create;

    _Files:=TStringList.Create;

    _state:=ssInitializing;
end;

destructor TRFilefind.Destroy;
begin
     inherited Destroy;
     if (Thread<>nil) then Stop;
     _Files.Free;
     _Directories.Free;
end;


procedure TRFilefind.Start;
var
    r: Integer;
    str: String;
begin
    Thread:=TFindThread.Create;

    Thread.Completed:=_onCompleted;
    Thread.DirChange:=_onChangeDir;
    Thread.Found:=_onFileFound;
    Thread.FoundText:=_onTextFound;
    Thread.Denied:=_onAccessDenied;
    Thread.Scanning:=_onStartScanning;
    Thread.filemask:=_filemask;
    Thread.Paths.Clear;
    Thread.Paths.AddStrings(_Directories);
    Thread.text:=_text;
    Thread.caseSensitive:=_caseSensitive;
    Thread.FreeOnTerminate:=FALSE;
    Thread.Priority:=_Threadpriority;
    Thread.Files.Clear;

    for r:=0 to _files.Count-1 do
    begin
        str:=_files.Strings[r];
        while (Length(str)>=1)and(str[1]='\') do
            str:=Copy(str,2,Length(str));
        if (Length(str)>=1) then Thread.Files.Add(str);
    end;

     _state:=ssRunning;
    Thread.Resume;
end;

procedure TRFilefind.Stop;
begin
    if(Thread<>nil) then
    begin
        Thread.Stop;
        _state:=ssStopped;
    end;
end;

procedure TRFilefind.Completed(Sender: TObject);
begin
    _state:=ssCompleted;
    if (@_OnCompleted<>nil) then _OnCompleted(Sender);
end;


procedure TRFilefind.Pause;
begin
    if (Thread<>nil)and(Thread.Suspended=FALSE) then
    begin
        _state:=ssPaused;
        Thread.Suspend;
    end;
end;

procedure TRFilefind.Continue;
begin
    if (Thread<>nil)and(Thread.Suspended=TRUE)and(_state=ssPaused) then
    begin
        _state:=ssRunning;
        Thread.Resume;
    end;
end;

procedure TRFilefind.Set_Text (str: String);
begin
   _text:=str;
end;

procedure TRFilefind.Set_Directories (list: TStrings);
var
   v: Integer;
begin
   _Directories.Clear;
   _Directories.AddStrings(list);

   repeat
       v:=_Directories.IndexOf('');
       if (v<>-1) then _Directories.Delete(v);
    until (v=-1);

   for v:=0 to _Directories.Count-1 do
   begin
        if (_Directories[v][length(_Directories[v])]<>'\') then
            _Directories[v]:=_Directories[v]+'\';
   end;
end;

procedure TRFilefind.Set_Files (list: TStrings);
begin
   _Files.Clear;
   _Files.AddStrings(list);
end;



function TRFilefind.GetRemoteDrives: String;
var
   r: Integer;
   buf: Array[0..4] of Char;
begin
   Result:='';
   for r:=Integer('A') to Integer('Z') do
   begin
        StrPCopy(@buf,Char(r)+':\');
        if (GetDriveType(@buf)=DRIVE_REMOTE) then
            Result:=Result+Char(r);
   end;
end;


function TRFilefind.GetRemovableDrives: String;
var
   r: Integer;
   buf: Array[0..4] of Char;
begin
   Result:='';
   for r:=Integer('A') to Integer('Z') do
   begin
        StrPCopy(@buf,Char(r)+':\');
        if (GetDriveType(@buf)=DRIVE_REMOVABLE) then
            Result:=Result+Char(r);
   end;
end;

function TRFilefind.GetFixedDrives: String;
var
   r: Integer;
   buf: Array[0..4] of Char;
begin
   Result:='';
   for r:=Integer('A') to Integer('Z') do
   begin
        StrPCopy(@buf,Char(r)+':\');
        if (GetDriveType(@buf)=DRIVE_FIXED) then
            Result:=Result+Char(r);
   end;
end;

function TRFilefind.GetCDRomDrives: String;
var
   r: Integer;
   buf: Array[0..4] of Char;
begin
   Result:='';
   for r:=Integer('A') to Integer('Z') do
   begin
        StrPCopy(@buf,Char(r)+':\');
        if (GetDriveType(@buf)=DRIVE_CDROM) then
            Result:=Result+Char(r);
   end;
end;

function TRFilefind.GetRamDrives: String;
var
   r: Integer;
   buf: Array[0..4] of Char;
begin
   Result:='';
   for r:=Integer('A') to Integer('Z') do
   begin
        StrPCopy(@buf,Char(r)+':\');
        if (GetDriveType(@buf)=DRIVE_RAMDISK) then
            Result:=Result+Char(r);
   end;
end;


procedure TRFilefind.Set_Version (str: String);
begin
    // Read-only property
end;


//---------FindThread------------------------------------------------------

constructor TFindThread.Create;
begin
     Paths:=TStringList.Create;
     Files:=TStringList.Create;
     inherited Create(TRUE);
end;

destructor TFindThread.Destroy;
begin
     inherited Destroy;
     Files.Free;
     Paths.Free;
end;

procedure TFindThread.Execute;
var
    n: Integer;
begin
    TextLength:=Length(text);

    if (CaseSensitive=FALSE) then Text:=LowerCase(Text);

    if (Paths.Count>=1) then
    begin
        for n:=0 to Paths.Count-1 do
        begin
            if (Terminated=FALSE) then
            begin
                RecursiveSearchDir (Paths[n])
            end;
        end;
    end;

    Synchronize(WhenCompleted);
end;

procedure TFindThread.RecursiveSearchDir(dir: String);
var
    rec: TSearchRec;
    v: Integer;
begin
     directory:=dir;
     Synchronize(WhenDirChange);

     if (Terminated=FALSE) then
     begin
         TestIfFileInDirectory(dir);

         if (Terminated=FALSE) then
         begin
             v:=FindFirst(dir+'\*',faDirectory,rec);
             if (v=0) then
             begin
                 if Boolean(rec.Attr and $10)and(rec.Name<>'.')and(rec.Name<>'..') then
                      RecursiveSearchDir (dir+rec.Name+'\');

                 repeat
                 v:=FindNext(rec);
                 if (v=0) and Boolean(rec.Attr and $10)and(rec.Name<>'.')and(rec.Name<>'..') then
                 begin
                     RecursiveSearchDir (dir+rec.Name+'\');
                 end;
                 until (v<>0)or(Terminated=TRUE);
                 FindClose(rec);
             end;
         end;
     end;
end;


procedure TFindThread.TestIfFileContainsText;
var
    h: Integer;
    size,sizeleft,total:Integer;
    buf: Array[0..bufsize] of Byte;
    r,n: Integer;
    Label Out;
begin
    continueProcessing:=TRUE;

    h:=FileOpen(Directory+'\'+foundRec.Name,fmOpenRead or fmShareDenyNone);
    if (h=-1) then
    begin
        Synchronize(WhenDenied);
        Exit;
    end;

    AllowProcessing:=TRUE;
    Synchronize(WhenScanning);
    if (AllowProcessing=FALSE) then Exit;

    sizeleft:=Foundrec.Size;
    total:=0;

    repeat
        if (sizeleft>bufsize) then size:=bufsize
            else size:=sizeleft;

        FileSeek(h,1,-textlength);    // Strings may appear in the gap between two blocks
        FileRead(h,buf,size);

        if (size<>sizeleft) then
            sizeleft:=sizeleft-size+textlength
        else sizeleft:=sizeleft-size;

        if (CaseSensitive=FALSE) then
        begin
            for r:=0 to size do
                if (buf[r]>=Byte('A'))and(buf[r]<=Byte('Z')) then
                    buf[r]:=buf[r]+32;
        end;


        for r:=0 to size-textlength do
        begin
            if (buf[r]=Byte(text[1])) then
            begin
                for n:=1 to textlength do
                begin
                    if (Byte(text[n])<>buf[r+n-1]) then Break;
                    if (n=textlength) then
                    begin
                        currentOffset:=total+r;
                        Synchronize(TextFound);
                        if (continueProcessing=FALSE) then Exit;
                    end;
                end;
            end;
        end;
        total:=total+size;
    until (sizeleft=0);

    FileClose(h);
end;


procedure TFindThread.TestIfFileInDirectory (dir: String);
var
    v: Integer;
    rec: TSearchRec;
begin
    if (Terminated=FALSE) then
    begin
        v:=FindFirst(dir+'\'+filemask,$3F,rec);
        if (v=0)then
        begin
            if(rec.Name<>'.')and(rec.Name<>'..') then
            begin
                FoundRec:=rec;
                Directory:=dir;
                if (Files.Count=0)or(Files.IndexOf(rec.Name)<>-1)or(Files.IndexOf(Directory+rec.Name)<>-1)or
                        (Files.IndexOf(Copy(Directory,4,255)+'\'+rec.Name)<>-1) then
                begin

                    if (textlength>=1) then
                    begin
                        if ((rec.Attr and faDirectory)=0)and(rec.Size>=textlength)then
                        begin
                             TestIfFileContainsText;
                        end;
                    end
                    else Synchronize(MatchFound);
                end;
            end;
            repeat
                v:=FindNext(rec);
                if (v=0)and(rec.Name<>'.')and(rec.Name<>'..') then
                begin
                    FoundRec:=rec;
                    Directory:=dir;
                    if (Files.Count=0)or(Files.IndexOf(rec.Name)<>-1)or(Files.IndexOf(Directory+rec.Name)<>-1)or
                        (Files.IndexOf(Copy(Directory,4,255)+'\'+rec.Name)<>-1) then
//                    if (Files.Count=0)or(Files.IndexOf(rec.Name)<>-1)or(Files.IndexOf(Directory+rec.Name)<>-1)or
//                        (Files.IndexOf(Copy(Directory,4,255)+rec.Name)<>-1) then
                    begin
                        if (textlength>=1) then
                        begin
                        if ((rec.Attr and faDirectory)=0)and(rec.Size>=textlength)then
                                TestIfFileContainsText;
                        end
                        else Synchronize(MatchFound);
                    end;
                end;
            until (v<>0);

            FindClose(rec);
        end;
    end;
end;

procedure TFindThread.Stop;
begin
    Terminate;
    Resume;
    while (Terminated=FALSE) do ;
end;

procedure TFindThread.WhenCompleted;
begin
    if (Terminated=FALSE) and
       (@Completed<>nil) then Completed(Self);
end;

procedure TFindThread.WhenDirChange;
begin
    if (Terminated=FALSE) then
        if (@DirChange<>nil) then
            DirChange(Self,directory);
end;

procedure TFindThread.MatchFound;
begin
    if (Terminated=FALSE) then
        if (@Found<>nil) then
           Found(Self,Directory,foundrec);
end;

procedure TFindThread.TextFound;
begin
    if (Terminated=FALSE) then
        if (@FoundText<>nil) then
           FoundText(Self,Directory,foundrec,currentOffset,continueProcessing);
end;

procedure TFindThread.WhenScanning;
begin
    if (Terminated=FALSE) then
        if (@Scanning<>nil) then
            Scanning(Self,Directory,foundrec,AllowProcessing);
end;

procedure TFindThread.WhenDenied;
begin
    if (@Denied<>nil) then Denied(Self,Directory,foundrec);
end;

end.





